//
// water.fx
//

//---------------------------------------------------------------------
// Water settings
//---------------------------------------------------------------------
texture showroomMap_Tex;
texture microflakeNMap_Tex;
float4 waterColor = float4(90 / 255.0, 170 / 255.0, 170 / 255.0, 240 / 255.0 );


//---------------------------------------------------------------------
// These parameters are set by MTA whenever a shader is drawn
//---------------------------------------------------------------------
float4x4 World;
float4x4 View;
float4x4 Projection;
float4x4 WorldViewProjection;

float3 cameraPosition : CAMERAPOSITION;

float gTime : TIME;


//------------------------------------------------------------------------------------------
// Samplers for the textures
//------------------------------------------------------------------------------------------
sampler microflakeNMap = sampler_state
{
   Texture = (microflakeNMap_Tex);
   MAGFILTER = LINEAR;
   MINFILTER = LINEAR;
   MIPFILTER = POINT;
   MIPMAPLODBIAS = 0.000000;
};

sampler showroomMap = sampler_state
{
   Texture = (showroomMap_Tex);
   MAGFILTER = LINEAR;
   MINFILTER = LINEAR;
   MIPFILTER = LINEAR;
   MIPMAPLODBIAS = 0.000000;
};


//---------------------------------------------------------------------
// Structure of data sent to the vertex and pixel shaders
//---------------------------------------------------------------------
struct VertexShaderInput
{
    float3 Position : POSITION0;
    float4 Diffuse : COLOR0;
    float2 TexCoord0 : TEXCOORD0;
};

struct PixelShaderInput
{
    float4 Pos  : POSITION;
    float4 Diffuse : COLOR0;
    float4 SparkleTex : TEXCOORD0;
    float3 WorldPosition : TEXCOORD1;
};


//------------------------------------------------------------------------------------------
// VertexShaderFunction
//------------------------------------------------------------------------------------------
PixelShaderInput VertexShaderFunction(VertexShaderInput In)
{
    PixelShaderInput Out = (PixelShaderInput)0;

    // Transform postion
    Out.Pos = mul(float4(In.Position, 1), WorldViewProjection);
    float3 worldPosition = mul(float4(In.Position, 1), (float4x3)World);
    float3 viewDirection = normalize(cameraPosition - worldPosition);

    // Transfer stuff
    Out.WorldPosition = worldPosition;

    // Scroll noise texture
    float2 uvpos1 = 0;
    float2 uvpos2 = 0;

    uvpos1.x = sin(gTime/40);
    uvpos1.y = fmod(gTime/50,1);

    uvpos2.x = fmod(gTime/10,1);
    uvpos2.y = sin(gTime/12);

    Out.SparkleTex.x = worldPosition.x / 6 + uvpos1.x;
    Out.SparkleTex.y = worldPosition.y / 6 + uvpos1.y;
    Out.SparkleTex.z = worldPosition.x / 10 + uvpos2.x;
    Out.SparkleTex.w = worldPosition.y / 10 + uvpos2.y;

    // Convert regular water color to what we want
    float4 waterColorBase = float4(90 / 255.0, 170 / 255.0, 170 / 255.0, 240 / 255.0 );
    float4 conv           = float4(30 / 255.0,  58 / 255.0,  58 / 255.0, 200 / 255.0 );
    Out.Diffuse = saturate( waterColor * conv / waterColorBase );

    return Out;
}




//------------------------------------------------------------------------------------------
// Temp hack
//------------------------------------------------------------------------------------------
// calculate face and coordinates for cubemap lookup
void indexCubeMap(float3 d, out int face, out float s, out float t)
{
        float3 absd;
        float sc, tc, ma;

        absd.x = abs(d.x);
        absd.y = abs(d.y);
        absd.z = abs(d.z);

        face = 0;

        if ((absd.x >= absd.y) && (absd.x >= absd.z)) {
                if (d.x > 0.0f) {
                        face = 0;
                        sc = -d.z; tc = -d.y; ma = absd.x;
                } else {
                        face = 1;
                        sc = d.z; tc = -d.y; ma = absd.x;
                }
        }

        if ((absd.y >= absd.x) && (absd.y >= absd.z)) {
                if (d.y > 0.0f) {
                        face = 2;
                        sc = d.x; tc = d.z; ma = absd.y;
                } else {
                        face = 3;
                        sc = d.x; tc = -d.z; ma = absd.y;
                }
        }

        if ((absd.z >= absd.x) && (absd.z >= absd.y)) {
                if (d.z > 0.0f) {
                        face = 4;
                        sc = d.x; tc = -d.y; ma = absd.z;
                } else {
                        face = 5;
                        sc = -d.x; tc = -d.y; ma = absd.z;
                }
        }

        if (ma == 0.0f) {
                s = 0.0f;
                t = 0.0f;
        } else {
                s = ((sc / ma) + 1.0f) * 0.5f;
                t = ((tc / ma) + 1.0f) * 0.5f;
        }
}

//------------------------------------------------------------------------------------------
// Also temp hack
//------------------------------------------------------------------------------------------
// lookup in cubemap stored in 2D texture
float4 mytexCUBE( float3 d )
{
    int face;
    float s, t;
    indexCubeMap(d, face, s, t);

    return tex2D(showroomMap, float2(s, t));
}


//------------------------------------------------------------------------------------------
// PixelShaderFunction
//------------------------------------------------------------------------------------------
float4 PixelShaderFunction(PixelShaderInput In) : COLOR0
{
    //
    // This was all ripped and modded from the car paint shader, so some of the comments may seem a bit strange
    //

    float brightnessFactor = 0.10;
    float glossLevel = 0.00;

    // Get the surface normal
    float3 vNormal = float3(0,0,1);

    // Micro-flakes normal map is a high frequency normalized
    // vector noise map which is repeated across the surface.
    // Fetching the value from it for each pixel allows us to
    // compute perturbed normal for the surface to simulate
    // appearance of micro-flakes suspended in the coat of paint:
    float3 vFlakesNormal = tex3D(microflakeNMap, float3(In.SparkleTex.xy,1)).rgb;
    float3 vFlakesNormal2 = tex3D(microflakeNMap, float3(In.SparkleTex.zw,1)).rgb;

    vFlakesNormal = (vFlakesNormal + vFlakesNormal2) / 2;

    // Don't forget to bias and scale to shift color into [-1.0, 1.0] range:
    vFlakesNormal = 2 * vFlakesNormal - 1.0;

    // To compute the surface normal for the second layer of micro-flakes, which
    // is shifted with respect to the first layer of micro-flakes, we use this formula:
    // Np2 = ( c * Np + d * N ) / || c * Np + d * N || where c == d
    float3 vNp2 = ( vFlakesNormal + vNormal ) ;

    // The view vector (which is currently in world space) needs to be normalized.
    // This vector is normalized in the pixel shader to ensure higher precision of
    // the resulting view vector. For this highly detailed visual effect normalizing
    // the view vector in the vertex shader and simply interpolating it is insufficient
    // and produces artifacts.
    float3 vView = normalize(cameraPosition - In.WorldPosition);

    // Transform the surface normal into world space (in order to compute reflection
    // vector to perform environment map look-up):
    float3 vNormalWorld = float3(0,0,1);

    // Compute reflection vector resulted from the clear coat of paint on the metallic
    // surface:
    float fNdotV = saturate(dot( vNormalWorld, vView));
    float3 vReflection = 2 * vNormalWorld * fNdotV - vView;

    // Hack in some bumpyness
    vReflection += vNp2;

    // Here we just use a constant gloss value to bias reading from the environment
    // map, however, in the real demo we use a gloss map which specifies which
    // regions will have reflection slightly blurred.

    float fEnvBias = 0;//fmod(gTime/4,4);
    // Sample environment map using this reflection vector:

    // float4 envMap = texCUBEbias( showroomMap, float4( vReflection, fEnvBias ) );
    float4 envMap = mytexCUBE( vReflection );
    // Premultiply by alpha:

    // Premultiply by alpha:
    envMap.rgb = envMap.rgb * envMap.a;

    // Brighten the environment map sampling result:
    envMap.rgb *= brightnessFactor;


    float4 OutColor = 1;

    // Bodge in the water color
    OutColor = envMap + In.Diffuse * 0.5;
    OutColor += envMap * In.Diffuse;
    OutColor.a = In.Diffuse.a;

    return OutColor;
}



//-----------------------------------------------------------------------------
// Techniques
//-----------------------------------------------------------------------------
technique water
{
    pass P0
    {
        VertexShader = compile vs_2_0 VertexShaderFunction();
        PixelShader  = compile ps_2_0 PixelShaderFunction();
    }
}


technique fallback
{
    pass P0
    {
    }
}
